'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.default = init;

var _uberproto = require('uberproto');

var _uberproto2 = _interopRequireDefault(_uberproto);

var _feathersQueryFilters = require('feathers-query-filters');

var _feathersQueryFilters2 = _interopRequireDefault(_feathersQueryFilters);

var _isPlainObject = require('is-plain-object');

var _isPlainObject2 = _interopRequireDefault(_isPlainObject);

var _errorHandler = require('./error-handler');

var _errorHandler2 = _interopRequireDefault(_errorHandler);

var _feathersErrors = require('feathers-errors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const METHODS = {
  $or: 'orWhere',
  $and: 'andWhere',
  $ne: 'whereNot',
  $in: 'whereIn',
  $nin: 'whereNotIn'
};

const OPERATORS = {
  $lt: '<',
  $lte: '<=',
  $gt: '>',
  $gte: '>=',
  $like: 'like'

  /**
   * Class representing an feathers adapter for objection.js ORM.
   * @param {object} options
   * @param {string} [id='id'] - database id field
   * @param {object} options.model - an objection model
   * @param {object} [options.paginate]
   * @param {string} [allowedEager] - Objection eager loading string.
   */
};class Service {
  constructor(options) {
    if (!options) {
      throw new Error('Objection options have to be provided');
    }

    if (!options.model) {
      throw new Error('You must provide an Objection Model');
    }

    this.options = options || {};
    this.id = options.id || 'id';
    this.paginate = options.paginate || {};
    this.events = options.events || [];
    this.Model = options.model;
    this.allowedEager = options.allowedEager || '[]';
    this.namedEagerFilters = options.namedEagerFilters;
    this.eagerFilters = options.eagerFilters;
  }

  extend(obj) {
    return _uberproto2.default.extend(obj, this);
  }

  /**
   * Maps a feathers query to the objection/knex schema builder functions.
   * @param query - a query object. i.e. { type: 'fish', age: { $lte: 5 }
   * @param params
   * @param parentKey
   */
  objectify(query, params, parentKey) {
    // Delete $eager
    if (params.$eager) {
      delete params.$eager;
    }

    // Delete $joinEager
    if (params.$joinEager) {
      delete params.$joinEager;
    }
    Object.keys(params || {}).forEach(key => {
      const value = params[key];

      if ((0, _isPlainObject2.default)(value)) {
        return this.objectify(query, value, key);
      }

      const column = parentKey || key;
      const method = METHODS[key];
      const operator = OPERATORS[key] || '=';

      if (method) {
        if (key === '$or' || key === '$and') {
          const self = this;

          return value.forEach(condition => {
            query[method](function () {
              self.objectify(this, condition);
            });
          });
        }

        return query[method].call(query, column, value); // eslint-disable-line no-useless-call
      }

      return operator === '=' ? query.where(column, value) : query.where(column, operator, value);
    });
  }

  _createQuery(params = {}) {
    let trx = params.transaction ? params.transaction.trx : null;
    let q = this.Model.query(trx);
    return q;
  }

  createQuery(params = {}) {
    const { filters, query } = (0, _feathersQueryFilters2.default)(params.query || {});
    let q = this._createQuery(params).skipUndefined().allowEager(this.allowedEager);

    // $select uses a specific find syntax, so it has to come first.
    if (filters.$select) {
      q = q.select(...filters.$select.concat(this.id));
    }

    // $eager for objection eager queries

    if (query && query.$eager) {
      q.eager(query.$eager, this.namedEagerFilters);
      delete query.$eager;
    }

    if (query && query.$joinEager) {
      q.eagerAlgorithm(this.Model.JoinEagerAlgorithm).eager(query.$joinEager, this.namedEagerFilters);
      delete query.$joinEager;
    }

    // apply eager filters if specified
    if (this.eagerFilters) {
      const eagerFilters = this.eagerFilters;
      if (Array.isArray(eagerFilters)) {
        for (var eagerFilter of eagerFilters) {
          q.filterEager(eagerFilter.expression, eagerFilter.filter);
        }
      } else {
        q.filterEager(eagerFilters.expression, eagerFilters.filter);
      }
    }

    // build up the knex query out of the query params
    this.objectify(q, query);

    if (filters.$sort) {
      Object.keys(filters.$sort).forEach(key => {
        q = q.orderBy(key, filters.$sort[key] === 1 ? 'asc' : 'desc');
      });
    }

    return q;
  }

  _find(params, count, getFilter = _feathersQueryFilters2.default) {
    const { filters, query } = getFilter(params.query || {});
    const q = params.objection || this.createQuery(params);

    // Handle $limit
    if (filters.$limit) {
      q.limit(filters.$limit);
    }

    // Handle $skip
    if (filters.$skip) {
      q.offset(filters.$skip);
    }

    let executeQuery = total => {
      return q.then(data => {
        return {
          total,
          limit: filters.$limit,
          skip: filters.$skip || 0,
          data
        };
      });
    };

    if (filters.$limit === 0) {
      executeQuery = total => {
        return Promise.resolve({
          total,
          limit: filters.$limit,
          skip: filters.$skip || 0,
          data: []
        });
      };
    }

    if (count) {
      let countQuery = this._createQuery(params).skipUndefined().count(`${this.id} as total`);

      this.objectify(countQuery, query);

      return countQuery.then(count => parseInt(count[0].total, 10)).then(executeQuery);
    }

    return executeQuery().catch(_errorHandler2.default);
  }

  /**
   * `find` service function for objection.
   * @param params
   */
  find(params) {
    const paginate = params && typeof params.paginate !== 'undefined' ? params.paginate : this.paginate;
    const result = this._find(params, !!paginate.default, query => (0, _feathersQueryFilters2.default)(query, paginate));

    if (!paginate.default) {
      return result.then(page => page.data);
    }

    return result;
  }

  _get(id, params) {
    const query = _extends({}, params.query);
    query[this.id] = id;

    return this._find(_extends({}, params, { query })).then(page => {
      if (page.data.length !== 1) {
        throw new _feathersErrors.errors.NotFound(`No record found for id '${id}'`);
      }

      return page.data[0];
    }).catch(_errorHandler2.default);
  }

  /**
   * `get` service function for objection.
   * @param {...object} args
   * @return {Promise} - promise containing the data being retrieved
   */
  get(...args) {
    return this._get(...args);
  }

  _create(data, params) {
    return this._createQuery(params).insert(data, this.id).then(row => {
      const id = typeof data[this.id] !== 'undefined' ? data[this.id] : row[this.id];
      return this._get(id, params);
    }).catch(_errorHandler2.default);
  }

  /**
   * `create` service function for objection.
   * @param {object} data
   * @param {object} params
   */
  create(data, params) {
    if (Array.isArray(data)) {
      return Promise.all(data.map(current => this._create(current, params)));
    }

    return this._create(data, params);
  }

  /**
   * `update` service function for objection.
   * @param id
   * @param data
   * @param params
   */
  update(id, data, params) {
    if (Array.isArray(data)) {
      return Promise.reject(new Error('Not replacing multiple records. Did you mean `patch`?'));
    }

    // NOTE (EK): First fetch the old record so
    // that we can fill any existing keys that the
    // client isn't updating with null;
    return this._get(id, params).then(oldData => {
      let newObject = {};

      for (var key of Object.keys(oldData)) {
        if (data[key] === undefined) {
          newObject[key] = null;
        } else {
          newObject[key] = data[key];
        }
      }

      // NOTE (EK): Delete id field so we don't update it
      delete newObject[this.id];

      return this._createQuery(params).where(this.id, id).update(newObject).then(() => {
        // NOTE (EK): Restore the id field so we can return it to the client
        newObject[this.id] = id;
        return newObject;
      });
    }).catch(_errorHandler2.default);
  }

  /**
   * `patch` service function for objection.
   * @param id
   * @param data
   * @param params
   */
  patch(id, raw, params) {
    const query = (0, _feathersQueryFilters2.default)(params.query || {}).query;
    const data = _extends({}, raw);

    const mapIds = page => page.data.map(current => current[this.id]);

    // By default we will just query for the one id. For multi patch
    // we create a list of the ids of all items that will be changed
    // to re-query them after the update
    const ids = id === null ? this._find(params).then(mapIds) : Promise.resolve([id]);

    if (id !== null) {
      query[this.id] = id;
    }

    let q = this._createQuery(params);

    this.objectify(q, query);

    delete data[this.id];

    return ids.then(idList => {
      // Create a new query that re-queries all ids that
      // were originally changed
      const findParams = _extends({}, params, {
        query: {
          [this.id]: { $in: idList },
          $select: params.query && params.query.$select
        }
      });

      return q.patch(data).then(() => {
        return this._find(findParams).then(page => {
          const items = page.data;

          if (id !== null) {
            if (items.length === 1) {
              return items[0];
            } else {
              throw new _feathersErrors.errors.NotFound(`No record found for id '${id}'`);
            }
          }

          return items;
        });
      });
    }).catch(_errorHandler2.default);
  }

  /**
   * `remove` service function for objection.
   * @param id
   * @param params
   */
  remove(id, params) {
    params.query = _extends({}, params.query);

    // NOTE (EK): First fetch the record so that we can return
    // it when we delete it.
    if (id !== null) {
      params.query[this.id] = id;
    }

    return this._find(params).then(page => {
      const items = page.data;
      const { query: queryParams } = (0, _feathersQueryFilters2.default)(params.query || {});
      const query = this._createQuery(params);

      this.objectify(query, queryParams);

      return query.delete().then(() => {
        if (id !== null) {
          if (items.length === 1) {
            return items[0];
          } else {
            throw new _feathersErrors.errors.NotFound(`No record found for id '${id}'`);
          }
        }

        return items;
      });
    }).catch(_errorHandler2.default);
  }
}

function init(options) {
  return new Service(options);
}

init.Service = Service;
module.exports = exports['default'];